/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.helpers;

import eu.cessda.pasc.osmhhandler.nesstar.models.cmmstudy.*;
import lombok.experimental.UtilityClass;
import org.jdom2.Element;

import java.util.Optional;
import java.util.function.Function;

import static eu.cessda.pasc.osmhhandler.nesstar.helpers.DocElementParser.*;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.HTMLFilter.CLEAN_CHARACTER_RETURNS_STRATEGY;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.HandlerConstants.*;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.OaiPmhConstants.*;

/**
 * Placeholder for various strategies to use to extract metadata for each field type
 *
 * @author moses AT doraventures DOT com

 */
@UtilityClass
class ParsingStrategies {

  static Function<Element, Optional<Country>> countryStrategyFunction() {
    return element -> {
      Country country = Country.builder()
              .iso2LetterCode(getAttributeValue(element, ABBR_ATTR).orElse(COUNTRY_NOT_AVAIL))
              .countryName(CLEAN_CHARACTER_RETURNS_STRATEGY.apply(element.getText()))
              .build();
      return Optional.of(country);
    };
  }

  static Function<Element, Optional<Pid>> pidStrategyFunction() {
    return element -> {
      var agencyBuilder = Pid.builder().pid(CLEAN_CHARACTER_RETURNS_STRATEGY.apply(element.getText()));
      getAttributeValue(element, AGENCY_ATTR).ifPresent(agencyBuilder::agency);
      return Optional.of(agencyBuilder.build());
    };
  }

  static Function<Element, Optional<String>> nullableElementValueStrategyFunction() {
    return element -> {
      String value = element.getText();
      return value.isEmpty() ? Optional.empty() : Optional.of(value);
    };
  }

  static Function<Element, Optional<String>> creatorStrategyFunction() {
    return element -> Optional.of(
            getAttributeValue(element, CREATOR_AFFILIATION_ATTR).map(
                    valueString -> (element.getText() + " (" + valueString + ")")).orElseGet(element::getText)
    );
  }

  static Function<Element, Publisher> publisherStrategyFunction() {
    return element -> Publisher.builder()
            .iso2LetterCode(getAttributeValue(element, ABBR_ATTR).orElse(PUBLISHER_NOT_AVAIL))
            .publisher(CLEAN_CHARACTER_RETURNS_STRATEGY.apply(element.getText())).build();
  }

  static Function<Element, Optional<TermVocabAttributes>> termVocabAttributeStrategyFunction(boolean hasControlledValue) {
    return element -> {
      Element concept = element.getChild(CONCEPT_EL, NESSTAR_NS);
      if (concept == null) {
        concept = new Element(EMPTY_EL);
      }
      return Optional.of(parseTermVocabAttrAndValues(element, concept, hasControlledValue));
    };
  }

  static Function<Element, String> uriStrategyFunction() {
    return element -> Optional.ofNullable(element.getAttributeValue(URI_ATTR)).orElse("");
  }

  static Function<Element, Optional<VocabAttributes>> samplingTermVocabAttributeStrategyFunction(boolean hasControlledValue) {

    return element -> {
      Optional<VocabAttributes> vocabValueAttrsOpt = Optional.empty();
      Optional<Element> concept = Optional.ofNullable(element.getChild(CONCEPT_EL, NESSTAR_NS));
      if (concept.isPresent()) {  //PUG req. only process if element has a <concept>
        Element conceptVal = concept.orElse(new Element(EMPTY_EL));
        VocabAttributes vocabValueAttrs = parseVocabAttrAndValues(element, conceptVal, hasControlledValue);
        vocabValueAttrsOpt = Optional.of(vocabValueAttrs);
      }
      return vocabValueAttrsOpt;
    };
  }

  static Function<Element, Optional<DataCollectionFreeText>> dataCollFreeTextStrategyFunction() {

    return element -> {

      // #243: Extract in all cases - previously if the @date attribute was present extraction was skipped
      if (!element.getText().isEmpty()) {
        var builder = DataCollectionFreeText.builder()
                .dataCollectionFreeText(element.getText());
        getAttributeValue(element, EVENT_ATTR).ifPresent(builder::event);
        return Optional.of(builder.build());
      }

      return Optional.empty();
    };
  }
}
