/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.service;

import eu.cessda.pasc.osmhhandler.nesstar.configuration.HandlerConfigurationProperties;
import eu.cessda.pasc.osmhhandler.nesstar.dao.GetRecordDoa;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.OaiPmhException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.XMLParseException;
import eu.cessda.pasc.osmhhandler.nesstar.helpers.OaiPmhHelpers;
import eu.cessda.pasc.osmhhandler.nesstar.models.cmmstudy.CMMStudy;
import eu.cessda.pasc.osmhhandler.nesstar.models.configuration.OaiPmh;
import eu.cessda.pasc.osmhhandler.nesstar.models.configuration.Repo;
import lombok.extern.slf4j.Slf4j;
import org.jdom2.Document;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.xpath.XPathFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.xml.XMLConstants;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;

import static eu.cessda.pasc.osmhhandler.nesstar.helpers.CMMStudyMapperUtil.*;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.RecordResponseValidator.validateResponse;

/**
 * Service Class responsible for querying the repository to fetch remote records.
 *
 * @author moses AT doraventures DOT com
 */
@Service
@Slf4j
public class GetRecordServiceImpl implements GetRecordService {

  private static final XPathFactory X_FACTORY = XPathFactory.instance();
  private final GetRecordDoa getRecordDoa;
  private final HandlerConfigurationProperties oaiPmhHandlerConfig;

  @Autowired
  public GetRecordServiceImpl(GetRecordDoa getRecordDoa, HandlerConfigurationProperties oaiPmhHandlerConfig) {
    this.getRecordDoa = getRecordDoa;
    this.oaiPmhHandlerConfig = oaiPmhHandlerConfig;
  }

  @Override
  public CMMStudy getRecord(String repositoryUrl, String studyIdentifier, String languageOverride) throws HandlerException {
    log.info("Querying Repo [{}] for StudyID [{}]", repositoryUrl, studyIdentifier);
    Repo repository = OaiPmhHelpers.getMetadataPrefix(repositoryUrl, oaiPmhHandlerConfig.getOaiPmh());
    if (languageOverride == null) {
      languageOverride = oaiPmhHandlerConfig.getOaiPmh().getMetadataParsingDefaultLang().getLang();
    }
    String fullUrl = OaiPmhHelpers.buildGetStudyFullUrl(repository, studyIdentifier);
    try (InputStream recordXML = getRecordDoa.getRecordXML(fullUrl)) {
      CMMStudy.CMMStudyBuilder builder = CMMStudy.builder();
      mapDDIRecordToCMMStudy(recordXML, builder, languageOverride);
      return builder.studyXmlSourceUrl(fullUrl).build();
    } catch (JDOMException | IOException e) {
      throw new XMLParseException(URI.create(fullUrl), e);
    }
  }

  private void mapDDIRecordToCMMStudy(InputStream recordXML, CMMStudy.CMMStudyBuilder builder, String defaultLanguage)
          throws JDOMException, IOException, OaiPmhException {

    OaiPmh oaiPmh = oaiPmhHandlerConfig.getOaiPmh();
    Document document = getSaxBuilder().build(recordXML);

    // We exit if the record has an <error> element
    validateResponse(document, X_FACTORY);

    // Short-Circuit. We carry on to parse beyond the headers only if record is active
    boolean isActiveRecord = parseHeaderElement(builder, document, X_FACTORY);
    if (isActiveRecord) {
      String defaultLangIsoCode = parseDefaultLanguage(document, X_FACTORY, defaultLanguage);
      parseStudyTitle(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseAbstract(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parsePidStudies(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseCreator(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseDataAccessFreeText(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseClassifications(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseKeywords(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseTypeOfTimeMethod(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseStudyAreaCountries(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseUnitTypes(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parsePublisher(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseYrOfPublication(builder, document, X_FACTORY);
      parseSamplingProcedureFreeTexts(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseTypeOfModeOfCollection(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseDataCollectionDates(builder, document, X_FACTORY);
      parseDataCollectionFreeTexts(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
      parseStudyUrl(builder, document, X_FACTORY, oaiPmh, defaultLangIsoCode);
    }
  }

  private SAXBuilder getSaxBuilder() {
    SAXBuilder saxBuilder = new SAXBuilder();
    saxBuilder.setProperty(XMLConstants.ACCESS_EXTERNAL_DTD, "");
    saxBuilder.setProperty(XMLConstants.ACCESS_EXTERNAL_SCHEMA, "");
    return saxBuilder;
  }
}
