/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.service;

import eu.cessda.pasc.osmhhandler.nesstar.configuration.HandlerConfigurationProperties;
import eu.cessda.pasc.osmhhandler.nesstar.dao.ListRecordHeadersDao;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.XMLParseException;
import eu.cessda.pasc.osmhhandler.nesstar.helpers.ListIdentifiersResponseValidator;
import eu.cessda.pasc.osmhhandler.nesstar.helpers.OaiPmhHelpers;
import eu.cessda.pasc.osmhhandler.nesstar.models.response.RecordHeader;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static eu.cessda.pasc.osmhhandler.nesstar.helpers.HandlerConstants.RECORD_HEADER;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.HandlerConstants.STUDY;
import static eu.cessda.pasc.osmhhandler.nesstar.helpers.OaiPmhConstants.*;

/**
 * Service implementation to handle Listing Record Headers.
 *
 * @author moses AT doraventures DOT com
 */
@Service
@Slf4j
public class ListRecordHeadersServiceImpl implements ListRecordHeadersService {

  private final ListRecordHeadersDao listRecordHeadersDao;
  private final DocumentBuilderFactory builderFactory;
  private final HandlerConfigurationProperties config;

  @Autowired
  public ListRecordHeadersServiceImpl(ListRecordHeadersDao listRecordHeadersDao,
                                      DocumentBuilderFactory builderFactory,
                                      HandlerConfigurationProperties config) {
    this.listRecordHeadersDao = listRecordHeadersDao;
    this.builderFactory = builderFactory;
    this.config = config;
  }

  private static List<RecordHeader> parseRecordHeaders(Document doc) {
    NodeList headers = doc.getElementsByTagName(HEADER_ELEMENT);
    return IntStream.range(0, headers.getLength()).mapToObj(headerRowIndex -> headers.item(headerRowIndex).getChildNodes())
            .map(ListRecordHeadersServiceImpl::parseRecordHeader)
            .collect(Collectors.toList());
  }

  private Document getDocument(String fullListRecordUrlPath) throws XMLParseException {
    try (InputStream recordHeadersXMLString = listRecordHeadersDao.listRecordHeaders(fullListRecordUrlPath)) {
      return builderFactory.newDocumentBuilder().parse(recordHeadersXMLString);
    } catch (SAXException | IOException | ParserConfigurationException e) {
      throw new XMLParseException(URI.create(fullListRecordUrlPath), e);
    }
  }

  private static Optional<String> parseResumptionToken(Document doc) {
    // OAI-PMH mandatory resumption tag in response.  Value can be empty to suggest end of list
    NodeList resumptionToken = doc.getElementsByTagName(RESUMPTION_TOKEN_ELEMENT);
    if (resumptionToken.getLength() > 0) {
      Node item = resumptionToken.item(0);
      if (!item.getTextContent().trim().isEmpty()) {
        return Optional.ofNullable(item.getTextContent());
      }
    }
    log.debug("ParseRecordHeaders: Resumption token empty.");
    return Optional.empty();
  }

  private static RecordHeader parseRecordHeader(NodeList headerElements) {

    RecordHeader.RecordHeaderBuilder recordHeaderBuilder = RecordHeader.builder();
    recordHeaderBuilder.recordType(RECORD_HEADER);

    for (int headerElementIndex = 0; headerElementIndex < headerElements.getLength(); headerElementIndex++) {
      String headerElementName = headerElements.item(headerElementIndex).getNodeName();
      String currentHeaderElementValue;

      switch (headerElementName) {
        case IDENTIFIER_ELEMENT:
          currentHeaderElementValue = headerElements.item(headerElementIndex).getTextContent();
          recordHeaderBuilder.identifier(currentHeaderElementValue);
          break;
        case DATESTAMP_ELEMENT:
          currentHeaderElementValue = headerElements.item(headerElementIndex).getTextContent();
          recordHeaderBuilder.lastModified(currentHeaderElementValue);
          break;
        default:
          // nothing to do
          break;
      }
      recordHeaderBuilder.type(STUDY); // Set All Nesstar identifiers as a study, to date only studies are there.
    }
    return recordHeaderBuilder.build();
  }

  @Override
  public List<RecordHeader> getRecordHeaders(String baseRepoUrl) throws HandlerException {

    var repoConfig = OaiPmhHelpers.getMetadataPrefix(baseRepoUrl, config.getOaiPmh());
    String fullListRecordUrlPath = OaiPmhHelpers.appendListRecordParams(repoConfig);
    Document doc = getDocument(fullListRecordUrlPath);

    // We exit if the response has an <error> element
    ListIdentifiersResponseValidator.validateResponse(doc);

    log.debug("Paring records headers for [{}].", baseRepoUrl);
    List<RecordHeader> recordHeaders = retrieveRecordHeaders(doc, baseRepoUrl);
    log.debug("ParseRecordHeaders ended for {}:  No more resumption tokens to process.", baseRepoUrl);

    return recordHeaders;
  }

  private List<RecordHeader> retrieveRecordHeaders(Document document, String baseRepoUrl) throws XMLParseException {

    Optional<String> resumptionToken;
    var recordHeaders = new ArrayList<RecordHeader>();

    do {
      // Parse and add all found record headers
      recordHeaders.addAll(parseRecordHeaders(document));

      // Check and loop when there is a resumptionToken
      resumptionToken = parseResumptionToken(document);
      if (resumptionToken.isPresent()) {
        String repoUrlWithResumptionToken = OaiPmhHelpers.appendListRecordResumptionToken(baseRepoUrl, resumptionToken.get());
        log.trace("Looping for [{}].", repoUrlWithResumptionToken);
        document = getDocument(repoUrlWithResumptionToken);
      }

    } while (resumptionToken.isPresent());

    return recordHeaders;
  }
}
