/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.controller;

import com.fasterxml.jackson.databind.ObjectMapper;
import eu.cessda.pasc.osmhhandler.nesstar.configuration.HandlerConfigurationProperties;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.service.APISupportedServiceImpl;
import eu.cessda.pasc.osmhhandler.nesstar.service.GetRecordService;
import org.hamcrest.core.IsNull;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.annotation.Import;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;

import static eu.cessda.pasc.osmhhandler.nesstar.mock.data.CMMStudyTestData.getCMMStudy;
import static org.mockito.BDDMockito.given;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;


/**
 * @author moses AT doraventures DOT com
 */
@RunWith(SpringRunner.class)
@WebMvcTest(GetRecordController.class)
@Import({APISupportedServiceImpl.class, HandlerConfigurationProperties.class})
@ActiveProfiles("test")
@WithMockUser(username = "user")
public class GetRecordControllerTest {

  private static final ObjectMapper MAPPER = new ObjectMapper();

  @Autowired
  private MockMvc mockMvc;

  @MockBean
  GetRecordService getRecordService;

  @Test
  public void shouldReturnRecordSuccessfully() throws Exception {

    // Given
      given(this.getRecordService.getRecord("http://kirkedata.nsd.uib.no", "StudyID222", null))
        .willReturn(getCMMStudy());

    String expectedCMMStudyJsonString = MAPPER.writeValueAsString(getCMMStudy());

    // When
    this.mockMvc.perform(get("/v0/GetRecord/CMMStudy/StudyID222?Repository=http://kirkedata.nsd.uib.no")
        .accept(MediaType.APPLICATION_JSON_VALUE))

        // Then
        .andExpect(status().isOk())
        .andExpect(status().reason(new IsNull<>()))
        .andExpect(content().json(expectedCMMStudyJsonString));
  }

  @Test
  public void shouldCatchAndBuildRelevantEntityMessageWhenACustomHandlerExceptionIsThrownReturn() throws Exception {

    // Given
    String studyIdentifier = "StudyID222";
    given(this.getRecordService.getRecord("http://kirkedata.nsd.uib.no", studyIdentifier, null)).willThrow(HandlerException.class);

    // When
    this.mockMvc.perform(get("/v0/GetRecord/CMMStudy/StudyID222?Repository=http://kirkedata.nsd.uib.no")
            .accept(MediaType.APPLICATION_JSON_VALUE))

            // Then
            .andExpect(status().isInternalServerError())
            .andExpect(status().reason(new IsNull<>()))
            .andExpect(content().json("{\"exception\": " + HandlerException.class.getName() + "}"));
  }

  @Test
  public void shouldCatchAndBuildRelevantEntityMessageWhenAnExceptionIsThrownReturn() throws Exception {

    // Given
    String studyIdentifier = "StudyID222";
    given(this.getRecordService.getRecord("http://kirkedata.nsd.uib.no", studyIdentifier, null)).willThrow(HandlerException.class);

    // When
    this.mockMvc.perform(get("/v0/GetRecord/CMMStudy/" + studyIdentifier + "?Repository=http://kirkedata.nsd.uib.no")
            .accept(MediaType.APPLICATION_JSON_VALUE))

            // Then
            .andExpect(status().isInternalServerError())
            .andExpect(status().reason(new IsNull<>()))
            .andExpect(content().json("{\"exception\": " + HandlerException.class.getName() + "}"));
  }

  @Test
  public void shouldReturnErrorMessageForNonSupportedAPIVersion() throws Exception {

    // Given
    String invalidVersion = "v99";

    // When
    this.mockMvc.perform(
            get("/" + invalidVersion + "/GetRecord/CMMStudy/StudyID222?Repository=http://kirkedata.nsd.uib.no")
                    .accept(MediaType.APPLICATION_JSON_VALUE))

            // Then
            .andExpect(status().isNotFound())
            .andExpect(status().reason(new IsNull<>()));
  }

  @Test
  public void shouldReturnErrorMessageForInvalidPathThatStartsWithANonSupportedAPIVersion() throws Exception {

    // Given
    String invalidVersion = "v99";

    // When
      this.mockMvc.perform(
              get("/" + invalidVersion + "/GetRecord/CMMInvalid/StudyID222/?Repository=http://kirkedata.nsd.uib.no")
                      .accept(MediaType.APPLICATION_JSON_VALUE))

              // Then
              .andExpect(status().isNotFound())
              .andExpect(status().reason(new IsNull<>()));
  }
}