/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.controller;

import com.fasterxml.jackson.databind.ObjectMapper;
import eu.cessda.pasc.osmhhandler.nesstar.configuration.HandlerConfigurationProperties;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.mock.data.RecordHeadersMock;
import eu.cessda.pasc.osmhhandler.nesstar.models.response.RecordHeader;
import eu.cessda.pasc.osmhhandler.nesstar.service.APISupportedServiceImpl;
import eu.cessda.pasc.osmhhandler.nesstar.service.ListRecordHeadersServiceImpl;
import org.hamcrest.core.IsNull;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.annotation.Import;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;

import java.util.List;

import static org.mockito.BDDMockito.given;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

/**
 * @author moses AT doraventures DOT com

 */
@RunWith(SpringRunner.class)
@WebMvcTest(ListRecordHeadersController.class)
@Import({APISupportedServiceImpl.class, HandlerConfigurationProperties.class})
@ActiveProfiles("test")
@WithMockUser(username = "user")
public class ListRecordHeadersControllerTest {

  @Autowired
  private MockMvc mockMvc;
  private static final ObjectMapper MAPPER = new ObjectMapper();

  @MockBean
  ListRecordHeadersServiceImpl listRecordHeadersService;

  @Test
  public void shouldReturnListRecordHeadersSuccessfully() throws Exception {

    // Given
    given(this.listRecordHeadersService.getRecordHeaders("http://kirkedata.nsd.uib.no"))
        .willReturn(RecordHeadersMock.getRecordHeaders());

    List<RecordHeader> expectedRecordHeaders = RecordHeadersMock.getRecordHeaders();
    String expectedRecordsJsonString = MAPPER.writeValueAsString(expectedRecordHeaders);

    // When
    this.mockMvc.perform(get("/v0/ListRecordHeaders?Repository=http://kirkedata.nsd.uib.no").accept(MediaType.APPLICATION_JSON_VALUE))

            // Then
            .andExpect(status().isOk())
            .andExpect(status().reason(new IsNull<>()))
            .andExpect(content().json(expectedRecordsJsonString));
  }

  @Test
  public void shouldReturnSystemErrorWhenAndExceptionIsThrownInternally() throws Exception {

    // Given
    given(this.listRecordHeadersService.getRecordHeaders("http://kirkedata.nsd.uib.no"))
            .willThrow(HandlerException.class);

    // When
    this.mockMvc.perform(get("/v0/ListRecordHeaders?Repository=http://kirkedata.nsd.uib.no").accept(MediaType.APPLICATION_JSON_VALUE))

            // Then
            .andExpect(status().isInternalServerError())
            .andExpect(status().reason(new IsNull<>()))
            .andExpect(content().json("{\"exception\":\"" + HandlerException.class.getName() + "\"}"));
  }

  @Test
  public void shouldReturnErrorMessageForNonSupportedAPIVersion() throws Exception {

    // When
      this.mockMvc.perform(get("/v122/ListRecordHeaders/?Repository=http://kirkedata.nsd.uib.no").accept(MediaType.APPLICATION_JSON_VALUE))

              // Then
              .andExpect(status().isNotFound())
              .andExpect(status().reason(new IsNull<>()));
  }

  @Test
  public void shouldReturnErrorMessageForInvalidPathThatStartsWithANonSupportedAPIVersion() throws Exception {

    // When
      this.mockMvc.perform(get("/v122/ListRecordHeaders/xxx?Repository=http://kirkedata.nsd.uib.no").accept(MediaType.APPLICATION_JSON_VALUE))

              // Then
              .andExpect(status().isNotFound())
              .andExpect(status().reason(new IsNull<>()));
  }
}