/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.dao;

import com.pgssoft.httpclient.HttpClientMock;
import com.pgssoft.httpclient.MockedServerResponse;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HTTPException;
import lombok.SneakyThrows;
import org.junit.Test;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;

import static org.assertj.core.api.BDDAssertions.then;

/**
 * Dao Spring Test class for retrieving record's xml
 *
 * @author moses AT doraventures DOT com

 */
public class GetRecordDoaImplTest {

    private final HttpClientMock httpClient = new HttpClientMock();

    @SneakyThrows
    private static void throwInterruptedException(MockedServerResponse.Builder responseBuilder) {
        throw new InterruptedException("Mocked");
    }

    @Test
    public void shouldReturnXMLPayloadOfGivenRecordIdentifierFromGivenRepoURL() throws IOException {

        // Given
        final String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        final String recordIdentifier = "http://fors-getdata.unil.ch:80/obj/fStudy/ch.sidos.ddi.468.7773";
        final String fullRecordUrl = repoUrl + "?verb=GetRecord&identifier=" + recordIdentifier + "&metadataPrefix=oai_ddi";

        httpClient.onGet(fullRecordUrl).doReturnXML(fullRecordUrl, StandardCharsets.UTF_8);

        // When
        GetRecordDoa recordDoa = new GetRecordDoaImpl(httpClient);
        try (InputStream responseXMLRecord = recordDoa.getRecordXML(fullRecordUrl)) {
            then(new String(responseXMLRecord.readAllBytes(), StandardCharsets.UTF_8)).isEqualTo(fullRecordUrl);
        }
    }

    @SuppressWarnings("resource")
    @Test(expected = HTTPException.class)
    public void shouldThrowExceptionWhenRemoteServerResponseIsNotSuccessful() throws IOException {

        // Given
        String expectedFullGetRecordUrl = "https://the.remote.endpoint/?verb=GetRecord&identifier=1683&metadataPrefix=ddi";

        httpClient.onGet(expectedFullGetRecordUrl).doReturn(500, expectedFullGetRecordUrl);

        // When
        GetRecordDoa recordDoa = new GetRecordDoaImpl(httpClient);
        recordDoa.getRecordXML(expectedFullGetRecordUrl);
    }

    @SuppressWarnings("resource")
    @Test(expected = IOException.class)
    public void shouldThrowExceptionWhenIOExceptionIsThrown() throws IOException {

        // Given
        String expectedFullGetRecordUrl = "https://the.remote.endpoint/?verb=GetRecord&identifier=1683&metadataPrefix=ddi";

        httpClient.onGet(expectedFullGetRecordUrl).doThrowException(new IOException());

        // When
        GetRecordDoa recordDoa = new GetRecordDoaImpl(httpClient);
        recordDoa.getRecordXML(expectedFullGetRecordUrl);
    }

    @SuppressWarnings("resource")
    @Test(expected = IllegalStateException.class)
    public void shouldThrowIllegalStateExceptionWhenInterrupted() throws IOException {

        // Given
        httpClient.onGet().doAction(GetRecordDoaImplTest::throwInterruptedException);

        // When
        GetRecordDoa recordDoa = new GetRecordDoaImpl(httpClient);
        recordDoa.getRecordXML("http://error.endpoint/");
    }
}
