/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.service;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.fge.jackson.JsonLoader;
import com.github.fge.jsonschema.core.exceptions.ProcessingException;
import com.github.fge.jsonschema.core.report.ProcessingReport;
import com.github.fge.jsonschema.main.JsonSchema;
import com.github.fge.jsonschema.main.JsonSchemaFactory;
import eu.cessda.pasc.osmhhandler.nesstar.dao.GetRecordDoa;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.OaiPmhException;
import eu.cessda.pasc.osmhhandler.nesstar.models.cmmstudy.CMMConverter;
import eu.cessda.pasc.osmhhandler.nesstar.models.cmmstudy.CMMStudy;
import lombok.extern.slf4j.Slf4j;
import org.json.JSONException;
import org.json.JSONObject;
import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringRunner;

import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import static eu.cessda.pasc.osmhhandler.nesstar.mock.data.CMMStudyTestData.*;
import static org.assertj.core.api.Assertions.fail;
import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.BDDMockito.given;
import static org.skyscreamer.jsonassert.JSONAssert.assertEquals;

/**
 * @author moses AT doraventures DOT com
 */
@RunWith(SpringRunner.class)
@SpringBootTest
@ActiveProfiles("test")
@Slf4j
public class GetRecordServiceImplTest {

  private final static String REPO_URL = "http://fors-nesstar.unil.ch/oai-pmh";
  private final static String STUDY_IDENTIFIER = "http://fors-getdata.unil.ch:80/obj/fStudy/ch.sidos.ddi.468.7773";
  private final static String FULL_RECORD_URL = REPO_URL + "?verb=GetRecord&identifier=" + STUDY_IDENTIFIER + "&metadataPrefix=oai_ddi";

  @MockBean
  GetRecordDoa getRecordDoa;

  @Autowired
  GetRecordService recordService;

  @Test
  public void shouldReturnValidCMMStudyRecordFromAFullyComplaintCmmDdiRecord() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL))
            .willReturn(getContentAsStream("xml/synthetic_compliant_cmm_nesstar.xml"));

    // When
    CMMStudy result = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(result).isNotNull();
    validateCMMStudyResultAgainstSchema(result);
    assertFieldsAreExtractedAsExpected(result);
  }

  @Test
  public void shouldReturnValidCMMStudyRecordFromOaiPmhDDI2_5MetadataRecord() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL))
            .willReturn(getContentAsStream("xml/synthetic_compliant_cmm_nesstar.xml"));

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(record).isNotNull();
    validateCMMStudyResultAgainstSchema(record);
  }

  @Test
  @SuppressWarnings("PreferJavaTimeOverload")
  public void shouldOnlyExtractSingleDateAsStartDateForRecordsWithASingleDateAttr() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL))
            .willReturn(getContentAsStream("xml/synthetic_compliant_cmm_nesstar_single_date.xml"));

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);
    then(record).isNotNull();
    validateCMMStudyResultAgainstSchema(record);
    final ObjectMapper mapper = new ObjectMapper();
    String jsonString = new CMMConverter().toJsonString(record);
    final JsonNode actualTree = mapper.readTree(jsonString);

    // Then
    then(actualTree.get("dataCollectionPeriodStartdate").asText()).isEqualTo("2004-09-16");
    then(actualTree.get("dataCollectionPeriodEnddate")).isNull();
    then(actualTree.get("dataCollectionYear").asInt()).isEqualTo(2004);
  }

  @Test
  public void shouldExtractDefaultLanguageFromCodebookXMLLagIfPresent() throws HandlerException, IOException {

    // Given
    InputStream expectedCmmStudyJsonString = getContentAsStream("xml/synthetic_compliant_cmm_nesstar.xml");
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL)).willReturn(expectedCmmStudyJsonString);

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(record.getTitleStudy().containsKey("xy")).isTrue();
    then(record.getAbstractField().containsKey("xy")).isTrue();
    then(record.getKeywords().containsKey("xy")).isTrue();
    then(record.getDataAccessFreeTexts().containsKey("xy")).isTrue();
  }

  @Test
  @SuppressWarnings("PreferJavaTimeOverload")
  public void shouldReturnCMMStudyRecordWithRepeatedAbstractConcatenated() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    Map<String, String> expectedAbstract = getAbstractFixture();
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL))
            .willReturn(getContentAsStream("xml/synthetic_compliant_cmm_nesstar_repeated_abstract.xml"));

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(record).isNotNull();
    then(record.getAbstractField().size()).isEqualTo(4);
    then(record.getAbstractField()).isEqualTo(expectedAbstract);
    validateCMMStudyResultAgainstSchema(record);
  }

  @Test // https://bitbucket.org/cessda/cessda.cdc.version2/issues/135
  @SuppressWarnings("PreferJavaTimeOverload")
  public void shouldReturnCMMStudyRecordWithOutParTitleWhenThereIsALangDifferentFromDefault() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    Map<String, String> expectedTitle = getTitleFixture();

    given(getRecordDoa.getRecordXML(FULL_RECORD_URL))
            .willReturn(getContentAsStream("xml/synthetic_compliant_cmm_nesstar_with_perTitl_xml_lang.xml"));

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(record).isNotNull();
    then(record.getTitleStudy().size()).isEqualTo(2);
    then(record.getTitleStudy()).isEqualTo(expectedTitle);
    validateCMMStudyResultAgainstSchema(record);
  }

  @Test()
  public void shouldReturnValidCMMStudyRecordFromOaiPmhDDI2_5MetadataRecord_MarkedAsNotActive()
          throws HandlerException, IOException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL)).willReturn(
            getContentAsStream("xml/synthetic_nesstar_deleted_record.xml")
    );

    // When
    CMMStudy record = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(record).isNotNull();
    then(record.isActive()).isFalse();
  }

  @Test(expected = OaiPmhException.class)
  public void shouldThrowExceptionForRecordWithErrorElement() throws HandlerException, IOException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL)).willReturn(
            getContentAsStream("xml/synthetic_nesstar_record_with_error.xml")
    );

    // When
    recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then --> an exception is thrown.
  }

  @Test
  public void shouldExtractAllRequiredCMMFieldsForAGivenAUKDSRecord() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL)).willReturn(
            getContentAsStream("xml/synthetic_compliant_cmm_nesstar.xml")
    );

    // When
    CMMStudy result = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, null);

    // Then
    then(result).isNotNull();
    validateCMMStudyResultAgainstSchema(result);
    assertThatCmmRequiredFieldsAreExtracted(result);
  }

  private void validateCMMStudyResultAgainstSchema(CMMStudy record) throws IOException, ProcessingException, JSONException {

    then(record.isActive()).isTrue(); // No need to carry on validating other fields if marked as inActive

    String jsonString = new CMMConverter().toJsonString(record);
    if (log.isDebugEnabled()) log.debug("RETRIEVED STUDY JSON: \n" + new JSONObject(jsonString).toString(4));

    JsonNode jsonNodeRecord = JsonLoader.fromString(jsonString);
    final JsonSchema schema = JsonSchemaFactory.byDefault().getJsonSchema("resource:/json/schema/CMMStudySchema.json");

    ProcessingReport validate = schema.validate(jsonNodeRecord);
    if (!validate.isSuccess()) {
      fail("Validation not successful: " + validate);
    }
  }

  @SuppressWarnings("PreferJavaTimeOverload")
  private void assertFieldsAreExtractedAsExpected(CMMStudy record) throws IOException, JSONException {

    final ObjectMapper mapper = new ObjectMapper();
    String jsonString = new CMMConverter().toJsonString(record);
    String expectedJson = getContent("json/synthetic_compliant_record_nesstar.json");
    final JsonNode actualTree = mapper.readTree(jsonString);
    final JsonNode expectedTree = mapper.readTree(expectedJson);

    // This following could be compared with one single Uber Json compare, but probably best this way to easily know
    // which field test assertion line below that fails.
    then(expectedTree.get("publicationYear").toString()).isEqualTo(actualTree.get("publicationYear").toString());
    then(expectedTree.get("dataCollectionPeriodStartdate").toString()).isEqualTo(actualTree.get("dataCollectionPeriodStartdate").toString());
    then(expectedTree.get("dataCollectionPeriodEnddate").toString()).isEqualTo(actualTree.get("dataCollectionPeriodEnddate").toString());
    then(expectedTree.get("dataCollectionYear").asInt()).isEqualTo(actualTree.get("dataCollectionYear").asInt());
    assertEquals(expectedTree.get("abstract").toString(), actualTree.get("abstract").toString(), true);
    assertEquals(expectedTree.get("classifications").toString(), actualTree.get("classifications").toString(), true);
    assertEquals(expectedTree.get("keywords").toString(), actualTree.get("keywords").toString(), true);
    assertEquals(expectedTree.get("typeOfTimeMethods").toString(), actualTree.get("typeOfTimeMethods").toString(), true);
    assertEquals(expectedTree.get("studyAreaCountries").toString(), actualTree.get("studyAreaCountries").toString(), true);
    assertEquals(expectedTree.get("pidStudies").toString(), actualTree.get("pidStudies").toString(), true);
    assertEquals(expectedTree.get("unitTypes").toString(), actualTree.get("unitTypes").toString(), true);
    assertEquals(expectedTree.get("titleStudy").toString(), actualTree.get("titleStudy").toString(), true);
    assertEquals(expectedTree.get("publisher").toString(), actualTree.get("publisher").toString(), true);
    assertEquals(expectedTree.get("creators").toString(), actualTree.get("creators").toString(), true);
    assertEquals(expectedTree
            .get("samplingProcedureFreeTexts").toString(), actualTree.get("samplingProcedureFreeTexts").toString(), true);
    assertEquals(expectedTree
            .get("typeOfModeOfCollections").toString(), actualTree.get("typeOfModeOfCollections").toString(), true);
    assertEquals(expectedTree
            .get("dataCollectionFreeTexts").toString(), actualTree.get("dataCollectionFreeTexts").toString(), true);
    assertEquals(expectedTree
            .get("dataAccessFreeTexts").toString(), actualTree.get("dataAccessFreeTexts").toString(), true);
    assertEquals(expectedTree.get("studyUrl").toString(), actualTree.get("studyUrl").toString(), true);
    assertEquals(expectedTree.get("studyXmlSourceUrl").toString(), actualTree.get("studyXmlSourceUrl").toString(), true);

    // n/a for nesstar
    //    assertEquals(expectedTree.get("fileLanguages").toString(), actualTree.get("fileLanguages").toString(), true);
    //    assertEquals(expectedTree.get("typeOfSamplingProcedures").toString(),
    //                 actualTree.get("typeOfSamplingProcedures").toString(), true);
  }

  private void assertThatCmmRequiredFieldsAreExtracted(CMMStudy record) throws IOException, JSONException {

    final ObjectMapper mapper = new ObjectMapper();
    String jsonString = new CMMConverter().toJsonString(record);
    String expectedJson = getContent("json/synthetic_compliant_record_nesstar.json");
    final JsonNode actualTree = mapper.readTree(jsonString);
    final JsonNode expectedTree = mapper.readTree(expectedJson);

    // CMM Model Schema required fields
    assertEquals(expectedTree.get("abstract").toString(), actualTree.get("abstract").toString(), true);
    assertEquals(expectedTree.get("titleStudy").toString(), actualTree.get("titleStudy").toString(), true);
    assertEquals(expectedTree.get("studyUrl").toString(), actualTree.get("studyUrl").toString(), true);
    then(expectedTree.get("studyNumber").toString()).isEqualTo(actualTree.get("studyNumber").toString());
    assertEquals(expectedTree.get("publisher").toString(), actualTree.get("publisher").toString(), true);
  }

  @Test
  public void shouldOverrideGlobalLanguageDefaultIfAPerRepositoryOverrideIsSpecified() throws HandlerException, IOException, ProcessingException, JSONException {

    // Given
    given(getRecordDoa.getRecordXML(FULL_RECORD_URL)).willReturn(
            getContentAsStream("xml/synthetic_compliant_cmm_nesstar_no_language.xml")
    );

    // When
    CMMStudy result = recordService.getRecord(REPO_URL, STUDY_IDENTIFIER, "zz");

    then(result).isNotNull();
    validateCMMStudyResultAgainstSchema(result);

    // Assert the language is as expected
    Assert.assertNotNull(result.getTitleStudy().get("zz"));
  }
}
