/*
 * Copyright © 2017-2021 CESSDA ERIC (support@cessda.eu)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package eu.cessda.pasc.osmhhandler.nesstar.service;

import eu.cessda.pasc.osmhhandler.nesstar.configuration.HandlerConfigurationProperties;
import eu.cessda.pasc.osmhhandler.nesstar.dao.ListRecordHeadersDao;
import eu.cessda.pasc.osmhhandler.nesstar.exception.HandlerException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.OaiPmhException;
import eu.cessda.pasc.osmhhandler.nesstar.exception.XMLParseException;
import eu.cessda.pasc.osmhhandler.nesstar.mock.data.RecordHeadersMock;
import eu.cessda.pasc.osmhhandler.nesstar.models.response.RecordHeader;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringRunner;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;

import static org.assertj.core.api.BDDAssertions.then;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;


/**
 * @author moses AT doraventures DOT com
 */
@RunWith(SpringRunner.class)
@SpringBootTest
@ActiveProfiles("test")
public class ListRecordHeadersServiceImplTest {

    private final ListRecordHeadersDao listRecordHeadersDao = mock(ListRecordHeadersDao.class);
    @Autowired
    private HandlerConfigurationProperties handlerConfigurationProperties;
    @Autowired
    private DocumentBuilderFactory documentBuilderFactory;

    @Test
    public void shouldReturnRecordHeadersForGivenRepo() throws HandlerException, IOException {

        // Given
        String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        String fullListRecordRepoUrl = "http://fors-nesstar.unil.ch/oai-pmh?verb=ListIdentifiers&metadataPrefix=oai_ddi&set=made_up_spec_val";
        String mockRecordHeadersXml = RecordHeadersMock.getListIdentifiersXMLResumptionEmpty();
        given(listRecordHeadersDao.listRecordHeaders(fullListRecordRepoUrl)).willReturn(new ByteArrayInputStream(mockRecordHeadersXml.getBytes(StandardCharsets.UTF_8)));

        ListRecordHeadersService listRecordHeadersService = new ListRecordHeadersServiceImpl(listRecordHeadersDao, documentBuilderFactory, handlerConfigurationProperties);

        // When
        List<RecordHeader> recordHeaders = listRecordHeadersService.getRecordHeaders(repoUrl);

        then(recordHeaders).hasSize(3);
        then(recordHeaders).extracting("identifier").containsOnly("850229", "850232", "850235");
        then(recordHeaders).extracting("lastModified").containsOnly("2017-11-20T10:37:18Z");
        then(recordHeaders).extracting("type").containsOnly("Study");
    }

    @Test(expected = IllegalArgumentException.class)
    public void shouldThrowWhenRequestForHeaderFails() throws HandlerException, IOException {

        // Given
        String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        String fullListRecordRepoUrl = "http://fors-nesstar.unil.ch/oai-pmh?verb=ListIdentifiers&metadataPrefix=oai_ddi&set=made_up_spec_val";
        String mockRecordHeadersXml = RecordHeadersMock.getListIdentifiersXMLResumptionTokenNotMockedForInvalid();
        given(listRecordHeadersDao.listRecordHeaders(fullListRecordRepoUrl)).willReturn(new ByteArrayInputStream(mockRecordHeadersXml.getBytes(StandardCharsets.UTF_8)));

        ListRecordHeadersService listRecordHeadersService = new ListRecordHeadersServiceImpl(listRecordHeadersDao, documentBuilderFactory, handlerConfigurationProperties);

        // When
        listRecordHeadersService.getRecordHeaders(repoUrl);
    }

    /**
     * FIXME: Currently non of the Nesstar repos are responding with Resumption tokens.  So removed resumption path
     * test until SPs comes back with a response to John's email regarding this.
     */
    @Test
    public void shouldRecursivelyLoopThroughTheOaiPMHResponseResumptionTokenToRetrieveReposCompleteListSize() throws HandlerException, IOException {

        // Given
        String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        String fullListRecordRepoUrl = "http://fors-nesstar.unil.ch/oai-pmh?verb=ListIdentifiers&metadataPrefix=oai_ddi&set=made_up_spec_val";
        String identifiersXML = RecordHeadersMock.getListIdentifiersXML();

        given(listRecordHeadersDao.listRecordHeaders(fullListRecordRepoUrl)).willReturn(new ByteArrayInputStream(identifiersXML.getBytes(StandardCharsets.UTF_8)));

        ListRecordHeadersService listRecordHeadersService = new ListRecordHeadersServiceImpl(listRecordHeadersDao, documentBuilderFactory, handlerConfigurationProperties);

        // When
        List<RecordHeader> recordHeaders = listRecordHeadersService.getRecordHeaders(repoUrl);

        then(recordHeaders).hasSize(3);
        then(recordHeaders).extracting("identifier")
                .containsOnly(
                        "http://fors-getdata.unil.ch:80/obj/fStudy/ch.sidos.ddi.391.7773",
                        "http://fors-getdata.unil.ch:80/obj/fStudy/ch.sidos.ddi.437.8163.98",
                        "http://fors-getdata.unil.ch:80/obj/fStudy/ch.sidos.ddi.437.8163.113");
        then(recordHeaders).extracting("lastModified")
                .containsOnly(
                        "2015-05-04T22:20:10+0000",
                        "2016-07-20T13:37:09+0000",
                        "2017-02-10T13:44:40+0000"
                );
        then(recordHeaders).extracting("type").containsOnly("Study");
    }

    @Test(expected = OaiPmhException.class)
    public void shouldThrowExceptionForRecordHeadersInvalidMetadataToken() throws HandlerException, IOException {

        // Given
        String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        String fullListRecordRepoUrl = "http://fors-nesstar.unil.ch/oai-pmh?verb=ListIdentifiers&metadataPrefix=oai_ddi&set=made_up_spec_val";
        String mockRecordHeadersXml = RecordHeadersMock.getListIdentifiersXMLWithInvalidMetadataTokenError();
        given(listRecordHeadersDao.listRecordHeaders(fullListRecordRepoUrl)).willReturn(new ByteArrayInputStream(mockRecordHeadersXml.getBytes(StandardCharsets.UTF_8)));

        ListRecordHeadersService listRecordHeadersService = new ListRecordHeadersServiceImpl(listRecordHeadersDao, documentBuilderFactory, handlerConfigurationProperties);

        // When
        listRecordHeadersService.getRecordHeaders(repoUrl);
    }

    @Test(expected = XMLParseException.class)
    public void shouldThrowOnIOException() throws HandlerException, ParserConfigurationException, IOException, SAXException {

        var documentBuilderFactoryMock = mock(DocumentBuilderFactory.class);
        var documentBuilder = mock(DocumentBuilder.class);

        // Given
        String repoUrl = "http://fors-nesstar.unil.ch/oai-pmh";
        String fullListRecordRepoUrl = "http://fors-nesstar.unil.ch/oai-pmh?verb=ListIdentifiers&metadataPrefix=oai_ddi&set=made_up_spec_val";
        var inputStream = InputStream.nullInputStream();
        given(listRecordHeadersDao.listRecordHeaders(fullListRecordRepoUrl)).willReturn(inputStream);

        when(documentBuilderFactoryMock.newDocumentBuilder()).thenReturn(documentBuilder);
        when(documentBuilder.parse(inputStream)).thenThrow(IOException.class);

        ListRecordHeadersService listRecordHeadersService = new ListRecordHeadersServiceImpl(listRecordHeadersDao, documentBuilderFactoryMock, handlerConfigurationProperties);

        // When
        listRecordHeadersService.getRecordHeaders(repoUrl);
    }
}
